class DumpTruck
  module Mysql
    class Translator
      TABLE_NAME_REGEX = /^CREATE TABLE `(.*)`/
      TABLE_FIELDS_REGEX = /^INSERT INTO `[^`]*` \((.*?)\)/
      TABLE_VALUES_REGEX = /VALUES (.*);/
      NUMBER = /-|\d/
      INTEGER = /-?\d+/
      FLOAT = /-?\d+\.\d+/
      NON_QUOTE = /([^'\\]|\\.)*/

      def table?(line)
        line =~ TABLE_NAME_REGEX
      end

      def extract_table(line)
        line.scan(TABLE_NAME_REGEX).flatten.first
      end

      def insert?(line)
        line =~ TABLE_FIELDS_REGEX
      end

      def extract_insert(line)
        fields = extract_fields(line)
        [fields, extract_rows(line, fields)]
      end

      def generate_insert(table, keys, data)
        "INSERT INTO `#{table}` (#{stringify_keys(keys)}) VALUES #{stringify_values(keys, data)};\n"
      end

    private

      def extract_fields(line)
        fields_string = line.scan(TABLE_FIELDS_REGEX).flatten.first
        fields_string.split(', ').map{|field| field[1..-2]}
      end

      def extract_rows(line, fields)
        values_string = line.scan(TABLE_VALUES_REGEX).flatten.first
        data = extract_values(values_string, fields, line)
        data.map{|values| Hash[fields.zip(values)]}
      end

      def extract_values(values_string, fields, line)
        scanner = StringScanner.new(values_string)
        data = []
        values = []

        until(scanner.eos?)
          values << case scanner.peek(1)
          when NUMBER
            extract_number(scanner)
          when "'"
            extract_string(scanner)
          when 'N'
            extract_null(scanner)
          when ')'
            scanner.getch
            data << values
            values = []
            next
          when /[(, ]/
            scanner.getch
            next
          else
            context = scanner.pre_match[-60..-1] + scanner.peek(1) + scanner.rest[0..60]
            raise "Unexpected character #{scanner.peek(1).inspect} at #{scanner.pos}. Last values: #{values.inspect}. Context: #{context.inspect}. Line: #{line.inspect}"
          end
        end

        data
      end

      def extract_number(scanner)
        if (number = scanner.scan(FLOAT))
          number.to_f
        else
          scanner.scan(INTEGER).to_i
        end
      end

      def extract_string(scanner)
        scanner.getch
        string = scanner.scan(NON_QUOTE)
        scanner.getch

        string
      end

      def extract_null(scanner)
        scanner.scan(/NULL/)
        nil
      end

      def stringify_keys(keys)
        keys.map{|key| "`#{key}`"}.join(', ')
      end

      def stringify_values(keys, data)
        data.map do |datum|
          values = datum.values_at(*keys)
          values_string = values.map{|value| stringify_field(value)}.join(',')
          "(#{values_string})"
        end.join(',')
      end

      def stringify_field(value)
        case value
        when nil
          'NULL'
        when Numeric
          value.to_s
        when String
          "'#{value}'"
        else
          raise "Unable to stringify #{value.inspect}"
        end
      end
    end
  end
end
